@@ -4,7 +4,7 @@ package JSON::Schema::Modern::Vocabulary::Content;
 # vim: set ts=8 sts=2 sw=2 tw=100 et :
 # ABSTRACT: Implementation of the JSON Schema Content vocabulary
 
-our $VERSION = '0.595';
+our $VERSION = '0.596';
 
 use 5.020;
 use Moo;
@@ -45,8 +45,6 @@ sub _traverse_keyword_contentEncoding ($class, $schema, $state) {
 sub _eval_keyword_contentEncoding ($class, $data, $schema, $state) {
   return 1 if not is_type('string', $data);
 
-  A($state, $schema->{$state->{keyword}});
-
   if ($state->{validate_content_schemas}) {
     my $decoder = $state->{evaluator}->get_encoding($schema->{contentEncoding});
     abort($state, 'cannot find decoder for contentEncoding "%s"', $schema->{contentEncoding})
@@ -62,6 +60,7 @@ sub _eval_keyword_contentEncoding ($class, $data, $schema, $state) {
     };
   }
 
+  A($state, $schema->{$state->{keyword}});
   return 1;
 }
 
@@ -70,8 +69,6 @@ sub _traverse_keyword_contentMediaType { goto \&_traverse_keyword_contentEncodin
 sub _eval_keyword_contentMediaType ($class, $data, $schema, $state) {
   return 1 if not is_type('string', $data);
 
-  A($state, $schema->{$state->{keyword}});
-
   if ($state->{validate_content_schemas}) {
     my $decoder = $state->{evaluator}->get_media_type($schema->{contentMediaType});
     abort($state, 'cannot find decoder for contentMediaType "%s"', $schema->{contentMediaType})
@@ -91,28 +88,27 @@ sub _eval_keyword_contentMediaType ($class, $data, $schema, $state) {
     }
   }
 
+  A($state, $schema->{$state->{keyword}});
   return 1;
 }
 
 sub _traverse_keyword_contentSchema ($class, $schema, $state) {
-  # since contentSchema should never be assumed to be evaluated in the context of the containing
-  # schema, it is not appropriate to gather identifiers found therein -- but we can still validate
-  # the subschema.
-  $class->traverse_subschema($schema, +{ %$state, identifiers => [] });
+  $class->traverse_subschema($schema, $state);
 }
 
 sub _eval_keyword_contentSchema ($class, $data, $schema, $state) {
   return 1 if not exists $schema->{contentMediaType};
   return 1 if not is_type('string', $data);
 
+  if ($state->{validate_content_schemas}) {
+    return 1 if not exists $state->{_content_ref};  # contentMediaType failed to decode the content
+    return E($state, 'subschema is not valid')
+      if not $class->eval($state->{_content_ref}->$*, $schema->{contentSchema},
+        { %$state, schema_path => $state->{schema_path}.'/contentSchema' });
+  }
+
   A($state, dclone($schema->{contentSchema}));
-  return 1 if not $state->{validate_content_schemas};
-
-  return 1 if not exists $state->{_content_ref};  # contentMediaType failed to decode the content
-
-  return 1 if $class->eval($state->{_content_ref}->$*, $schema->{contentSchema},
-    { %$state, schema_path => $state->{schema_path}.'/contentSchema' });
-  return E($state, 'subschema is not valid');
+  return 1;
 }
 
 1;
@@ -129,7 +125,7 @@ JSON::Schema::Modern::Vocabulary::Content - Implementation of the JSON Schema Co
 
 =head1 VERSION
 
-version 0.595
+version 0.596
 
 =head1 DESCRIPTION
 
